﻿/* Copyright (c) 2012 Rick (rick 'at' gibbed 'dot' us)
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would
 *    be appreciated but is not required.
 * 
 * 2. Altered source versions must be plainly marked as such, and must not
 *    be misrepresented as being the original software.
 * 
 * 3. This notice may not be removed or altered from any source
 *    distribution.
 */

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using Gibbed.Duels.FileFormats;
using Gibbed.Squish;
using NDesk.Options;
using Tdx = Gibbed.Duels.FileFormats.Tdx;

namespace Gibbed.Duels.TdxConvert
{
    internal class Program
    {
        private static string GetExecutableName()
        {
            return Path.GetFileName(System.Reflection.Assembly.GetExecutingAssembly().Location);
        }

        public static void Main(string[] args)
        {
            bool showHelp = false;
            bool saveExtra = false;
			bool compressDxt1 = false;
			bool compressDxt3 = false;
			bool compressDxt5 = false;
			bool nomips = false;

            var options = new OptionSet()
            {
				{
					"d1|dxt1",
					"when converting to TDX compress using DXT1 compression",
					v => compressDxt1 = v != null
					},
				{
					"d3|dxt3",
					"when converting to TDX compress using DXT3 compression",
					v => compressDxt3 = v != null
					},
				{
					"d5|dxt5",
					"when converting to TDX compress using DXT5 compression",
					v => compressDxt5 = v != null
					},
                {
                    "e|extra",
                    "save extra data if present when converting from TDX",
                    v => saveExtra = v != null
                    },
				{
					"nomip",
					"do not create mipmaps when compressing to TDX.",
					v => nomips = v != null
					},
                {
                    "h|help",
                    "show this message and exit",
                    v => showHelp = v != null
                    },
            };

            List<string> extra;

            try
            {
                extra = options.Parse(args);
            }
            catch (OptionException e)
            {
                Console.Write("{0}: ", GetExecutableName());
                Console.WriteLine(e.Message);
                Console.WriteLine("Try `{0} --help' for more information.", GetExecutableName());
                return;
            }

            if (extra.Count < 1 || showHelp == true)
            {
                Console.WriteLine("Usage: {0} [OPTIONS]+ input_texture*", GetExecutableName());
                Console.WriteLine("Convert a texture to / from Duels of the Planewalkers TDX format.");
                Console.WriteLine();
                Console.WriteLine("Options:");
                options.WriteOptionDescriptions(Console.Out);
                return;
            }

            foreach (string inputPath in extra)
            {
                Console.WriteLine(inputPath);

                var extension = Path.GetExtension(inputPath);
                if (extension != null)
                {
                    extension = extension.ToLowerInvariant();
                }

                // convert from tdx
                if (extension == ".tdx")
                {
                    Stream input = File.OpenRead(inputPath);
                    var tdx = new TdxFile();
                    tdx.Deserialize(input);
                    input.Close();

                    if (saveExtra == true && tdx.ExtraData != null)
                    {
                        using (var output = File.Create(Path.ChangeExtension(inputPath, "_extra.bin")))
                        {
                            output.Write(tdx.ExtraData, 0, tdx.ExtraData.Length);
                        }
                    }

                    if (tdx.Format == Tdx.D3DFormat.DXT1 ||
                        tdx.Format == Tdx.D3DFormat.DXT3 ||
                        tdx.Format == Tdx.D3DFormat.DXT5)
                    {
                        Native.Flags flags = 0;

                        if (tdx.Format == Tdx.D3DFormat.DXT1)
                        {
                            flags |= Native.Flags.DXT1;
                        }
                        else if (tdx.Format == Tdx.D3DFormat.DXT3)
                        {
                            flags |= Native.Flags.DXT3;
                        }
                        else if (tdx.Format == Tdx.D3DFormat.DXT5)
                        {
                            flags |= Native.Flags.DXT5;
                        }

                        var decompressed = Native.DecompressImage(
                            tdx.Mipmaps[0].Data,
                            tdx.Mipmaps[0].Width,
                            tdx.Mipmaps[0].Height,
                            flags);

                        var bitmap = MakeBitmapFromDXT(
                            tdx.Mipmaps[0].Width,
                            tdx.Mipmaps[0].Height,
                            decompressed,
                            true);

                        bitmap.Save(
                            Path.ChangeExtension(inputPath, ".png"),
                            ImageFormat.Png);
					}
                    else if (tdx.Format == Tdx.D3DFormat.A8R8G8B8)
                    {
                        var bitmap = MakeBitmapFromA8R8G8B8(
                            tdx.Mipmaps[0].Width,
                            tdx.Mipmaps[0].Height,
                            tdx.Mipmaps[0].Data);

                        bitmap.Save(
                            Path.ChangeExtension(inputPath, ".png"),
                            ImageFormat.Png);
                    }
                    else if (tdx.Format == Tdx.D3DFormat.A4R4G4B4)
                    {
                        var bitmap = MakeBitmapFromA4R4G4B4(
                            tdx.Mipmaps[0].Width,
                            tdx.Mipmaps[0].Height,
                            tdx.Mipmaps[0].Data);

                        bitmap.Save(
                            Path.ChangeExtension(inputPath, ".png"),
                            ImageFormat.Png);
                    }
                    else if (tdx.Format == Tdx.D3DFormat.X8R8G8B8)
                    {
                        var bitmap = MakeBitmapFromX8R8G8B8(
                            tdx.Mipmaps[0].Width,
                            tdx.Mipmaps[0].Height,
                            tdx.Mipmaps[0].Data);

                        bitmap.Save(
                            Path.ChangeExtension(inputPath, ".png"),
                            ImageFormat.Png);
                    }
                    else
                    {
                        throw new NotSupportedException("unsupported format " + tdx.Format.ToString());
                    }
                }
                    // convert to tdx
                else
                {
                    // just a crappy convert to A8R8G8B8 for now...
					// Riiak Shi Nal: Added capability to output to DXT1, DXT3, and DXT5.

                    using (var bitmap = new Bitmap(inputPath))
                    {
						Size szClosestMoFSize = FindClosestMultipleOf4Size(bitmap.Size);
						bool bCompress = compressDxt1 || compressDxt3 || compressDxt5;

                        var tdx = new TdxFile();
						if (bCompress)
						{
							tdx.Width = (ushort)szClosestMoFSize.Width;
							tdx.Height = (ushort)szClosestMoFSize.Height;
						}
						else
						{
							tdx.Width = (ushort)bitmap.Width;
							tdx.Height = (ushort)bitmap.Height;
						}
                        tdx.Flags = 0;
						Native.Flags eFlags = Native.Flags.ColourIterativeClusterFit;
						if (compressDxt1)
						{
							tdx.Format = Tdx.D3DFormat.DXT1;
							eFlags |= Native.Flags.DXT1;
						}
						else if (compressDxt3)
						{
							tdx.Format = Tdx.D3DFormat.DXT3;
							eFlags |= Native.Flags.DXT3;
						}
						else if (compressDxt5)
						{
							tdx.Format = Tdx.D3DFormat.DXT5;
							eFlags |= Native.Flags.DXT5;
						}
						else
							tdx.Format = Tdx.D3DFormat.A8R8G8B8;

						// Add our main image (top level).
						AddMipMap(tdx, bitmap, tdx.Width, tdx.Height, eFlags, bCompress);

						// Add MipMaps unless the user said not to.
						if ((bCompress) && (!nomips))
						{
							// Add the lower levels.
							ushort usLevelWidth = tdx.Width;
							ushort usLevelHeight = tdx.Height;
							while ((usLevelWidth > 1) || (usLevelHeight > 1))
							{
								// Reduce size and keep to even sizing (except when a dimension is less than 2 then that dimension is 1).
								//  Even sizing removed because the TDX format and game want images that are divided by two and truncated to a whole integer (was causing corrupted image appearance).
								usLevelWidth /= 2;
								//usLevelWidth -= (ushort)(usLevelWidth % 2);
								if (usLevelWidth < 1)
									usLevelWidth = 1;
								usLevelHeight /= 2;
								//usLevelHeight -= (ushort)(usLevelHeight % 2);
								if (usLevelHeight < 1)
									usLevelHeight = 1;
								AddMipMap(tdx, bitmap, usLevelWidth, usLevelHeight, eFlags, bCompress);
							}
						}

						/*var mip = new Tdx.Mipmap();
						mip.Width = tdx.Width;
						mip.Height = tdx.Height;

						int stride = mip.Width * 4;

						var area = new Rectangle(0, 0, bitmap.Width, bitmap.Height);
						var buffer = new byte[mip.Height * stride];

						var bitmapData = bitmap.LockBits(area, ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
						IntPtr scan = bitmapData.Scan0;
						for (int y = 0, o = 0; y < mip.Height; y++, o += stride)
						{
							Marshal.Copy(scan, buffer, o, stride);
							scan += bitmapData.Stride;
						}
						bitmap.UnlockBits(bitmapData);

						if ((tdx.Format == Tdx.D3DFormat.DXT1) ||
							(tdx.Format == Tdx.D3DFormat.DXT3) ||
							(tdx.Format == Tdx.D3DFormat.DXT5))
						{
							for (uint i = 0; i < mip.Width * mip.Height * 4; i += 4)
							{
								// flip red and blue
								byte r = buffer[i + 0];
								buffer[i + 0] = buffer[i + 2];
								buffer[i + 2] = r;
							}
							int newImageSize = tdx.GetMipSize(tdx.Format, tdx.Width, tdx.Height);
							if (tdx.Format == Tdx.D3DFormat.DXT1)
								mip.Data = Native.CompressImage(buffer, tdx.Width, tdx.Height, Native.Flags.DXT1 | Native.Flags.ColourIterativeClusterFit, newImageSize);
							else if (tdx.Format == Tdx.D3DFormat.DXT3)
								mip.Data = Native.CompressImage(buffer, tdx.Width, tdx.Height, Native.Flags.DXT3 | Native.Flags.ColourIterativeClusterFit, newImageSize);
							else if (tdx.Format == Tdx.D3DFormat.DXT5)
								mip.Data = Native.CompressImage(buffer, tdx.Width, tdx.Height, Native.Flags.DXT5 | Native.Flags.ColourIterativeClusterFit, newImageSize);
						}
						else
							mip.Data = buffer;

						tdx.Mipmaps.Add(mip);//*/

                        using (var output = File.Create(Path.ChangeExtension(inputPath, ".TDX")))
                        {
                            tdx.Serialize(output);
                        }
                    }
                }
            }
        }

		// Riiak Shi Nal: Added to automatically adjust to closest MoF size.
		public static Size FindClosestMultipleOf4Size(Size szInput)
		{
			double dRatio = ((double)szInput.Height) / ((double)szInput.Width);
			int nDistance = 0;

			if (((szInput.Width % 4) != 0) || ((szInput.Height % 4) != 0))
			{
				// Try going to the closest larger Multiple of Four and see if that works
				int nWidthMof = szInput.Width + (4 - (szInput.Width % 4));
				Size szReturn = new Size(nWidthMof, (int)Math.Round(nWidthMof * dRatio));

				// While we have a non-multiple of 4 size keep looking for one.
				while ((((szReturn.Width % 4) != 0) || ((szReturn.Height % 4) != 0)) && (nDistance < 50))
				{
					nDistance++;
					// Try enlarging by our distance to see if we can get a Mof size.
					szReturn.Width = nWidthMof + (4 * nDistance);
					szReturn.Height = (int)Math.Round(szReturn.Width * dRatio);
					if (((szReturn.Width % 4) == 0) && ((szReturn.Height % 4) == 0))
						break;
					// No luck enlarging so lets try reducing (as long as it would remain above 0.
					if ((nWidthMof - (4 * nDistance)) > 0)
					{
						szReturn.Width = nWidthMof - (4 * nDistance);
						szReturn.Height = (int)Math.Round(szReturn.Width * dRatio);
					}
					// loop will do our checking for us so no need for an extra if.
				}

				if (((szReturn.Width % 4) != 0) || ((szReturn.Height % 4) != 0))
				{
					// We must have hit our maximum distance without finding a Mof match (should never happen) now we need to go to our fallback (512x512).
					szReturn = new Size(512, 512);
				}

				return szReturn;
			}
			else
				return szInput;
		}

		// Riiak Shi Nal: Added to make creating mipmaps easier.
		private static void AddMipMap(TdxFile tdx, Bitmap bmpImage, ushort usWidth, ushort usHeight, Native.Flags eFlags, bool bCompress)
		{
			Tdx.Mipmap mip = new Tdx.Mipmap();
			mip.Width = usWidth;
			mip.Height = usHeight;

			int stride = mip.Width * 4;

			if ((bmpImage.Width != usWidth) || (bmpImage.Height != usHeight))
			{
				Size szNewSize = new Size(usWidth, usHeight);
				Bitmap bmpNew = new Bitmap(szNewSize.Width, szNewSize.Height);
				Graphics grfx = Graphics.FromImage(bmpNew);
				grfx.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
				grfx.DrawImage(bmpImage, 0, 0, szNewSize.Width, szNewSize.Height);
				bmpImage = bmpNew;
			}

			Rectangle area = new Rectangle(0, 0, usWidth, usHeight);
			byte[] buffer = new byte[mip.Height * stride];

			BitmapData bitmapData = bmpImage.LockBits(area, ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
			IntPtr scan = bitmapData.Scan0;
			for (int y = 0, o = 0; y < mip.Height; y++, o += stride)
			{
				Marshal.Copy(scan, buffer, o, stride);
				scan += bitmapData.Stride;
			}
			bmpImage.UnlockBits(bitmapData);

			if (bCompress)
			{
				for (uint i = 0; i < mip.Width * mip.Height * 4; i += 4)
				{
					// flip red and blue
					byte r = buffer[i + 0];
					buffer[i + 0] = buffer[i + 2];
					buffer[i + 2] = r;
				}
				int nNewImageSize = tdx.GetMipSize(tdx.Format, usWidth, usHeight);
				mip.Data = Native.CompressImage(buffer, usWidth, usHeight, eFlags, nNewImageSize);
			}
			else
				mip.Data = buffer;

			tdx.Mipmaps.Add(mip);
		}

		private static Bitmap MakeBitmapFromDXT(uint width, uint height, byte[] buffer, bool keepAlpha)
        {
            Bitmap bitmap = new Bitmap((int)width, (int)height, PixelFormat.Format32bppArgb);

            for (uint i = 0; i < width * height * 4; i += 4)
            {
                // flip red and blue
                byte r = buffer[i + 0];
                buffer[i + 0] = buffer[i + 2];
                buffer[i + 2] = r;
            }

            Rectangle area = new Rectangle(0, 0, bitmap.Width, bitmap.Height);
            BitmapData data = bitmap.LockBits(area, ImageLockMode.WriteOnly, bitmap.PixelFormat);
            Marshal.Copy(buffer, 0, data.Scan0, (int)(width * height * 4));
            bitmap.UnlockBits(data);
            return bitmap;
        }

        private static Bitmap MakeBitmapFromA8R8G8B8(uint width, uint height, byte[] buffer)
        {
            Bitmap bitmap = new Bitmap((int)width, (int)height, PixelFormat.Format32bppArgb);
            Rectangle area = new Rectangle(0, 0, bitmap.Width, bitmap.Height);
            BitmapData data = bitmap.LockBits(area, ImageLockMode.WriteOnly, bitmap.PixelFormat);
            Marshal.Copy(buffer, 0, data.Scan0, (int)(width * height * 4));
            bitmap.UnlockBits(data);
            return bitmap;
        }

        private static Bitmap MakeBitmapFromA4R4G4B4(uint width, uint height, byte[] buffer)
        {
            Bitmap bitmap = new Bitmap((int)width, (int)height, PixelFormat.Format32bppArgb);

            byte[] newbuffer = new byte[width * height * 4];

            for (uint i = 0, j = 0; i < width * height * 2; i += 2, j += 4)
            {
                newbuffer[j + 0] = (byte)(((buffer[i + 0] >> 0) & 0x0F) * 0x11); // A
                newbuffer[j + 1] = (byte)(((buffer[i + 0] >> 4) & 0x0F) * 0x11); // R
                newbuffer[j + 2] = (byte)(((buffer[i + 1] >> 0) & 0x0F) * 0x11); // G
                newbuffer[j + 3] = (byte)(((buffer[i + 1] >> 4) & 0x0F) * 0x11); // B
            }

            Rectangle area = new Rectangle(0, 0, bitmap.Width, bitmap.Height);
            BitmapData data = bitmap.LockBits(area, ImageLockMode.WriteOnly, bitmap.PixelFormat);
            Marshal.Copy(newbuffer, 0, data.Scan0, (int)(width * height * 4));
            bitmap.UnlockBits(data);
            return bitmap;
        }

        private static Bitmap MakeBitmapFromX8R8G8B8(uint width, uint height, byte[] buffer)
        {
            Bitmap bitmap = new Bitmap((int)width, (int)height, PixelFormat.Format32bppRgb);
            Rectangle area = new Rectangle(0, 0, bitmap.Width, bitmap.Height);
            BitmapData data = bitmap.LockBits(area, ImageLockMode.WriteOnly, bitmap.PixelFormat);
            Marshal.Copy(buffer, 0, data.Scan0, (int)(width * height * 4));
            bitmap.UnlockBits(data);
            return bitmap;
        }
    }
}
